
import requests
from config import USER_NOT_EXIST, URL_TIMEOUT, COMMON_ERROR, STATE_ERROR_LIST, SUCESSFUL_STATE
from tools import Logger


class CrawlerBase(object):
    # 初始化
    def __init__(self):
        pass
    # 获取请求体， 父类获取方法

    @staticmethod
    def get_request_body(url=None, headers=None, username=None):
        if username is not None:
            url = url + username
        Logger.common("get url: {}".format(url))
        response = None
        result = None
        try:
            response = requests.get(url=url, headers=headers, allow_redirects=False, timeout=10)
            response.raise_for_status()
            Logger.common("{}: request status :{}".format(username, response.status_code))
            if response.status_code == 200:
                result = response.content.decode("utf-8")
            elif response.status_code == 404:
                result = USER_NOT_EXIST
            # codeforces 用户不存在会重定向
            elif response.status_code == 302:
                result = USER_NOT_EXIST
            elif response.status_code == 408:
                result = URL_TIMEOUT
            elif response.status_code == 404:
                result = USER_NOT_EXIST
            else:
                result = COMMON_ERROR
        except requests.exceptions.ConnectionError:
            Logger.error("user->{} ConnectionError".format(username))
        except requests.exceptions.URLRequired:
            Logger.error("user->{} URLRequired".format(username))
        except requests.exceptions.Timeout:
            Logger.error("user->{} request TimeoutError".format(username))
        except requests.exceptions.InvalidURL:
            Logger.error("user->{} request InvalidURL".format(username))
        except Exception:
            Logger.error("user->{} unknow error")
        finally:
            return result

    def check_result(*result):
        """
            返回传递参数是否存在于error状态
            return bool -> True:状态ok
        """
        for item in result:
            if item in STATE_ERROR_LIST:
                return False
        return True





import json


class ConfigBase(object):
    """
        获取配置的基类，该基类通过注解，把配置文件注入相应的爬取分类中。
    """
    NAME_CONFIG_DICT = {}
    # 工厂模式， 执行读取config之后的函数

    @classmethod
    def register(cls, config_name=None):
        def wrapper(func):
            if func.__name__ not in cls.NAME_CONFIG_DICT.keys():
                with open("./config/{}.json".format(config_name)) as f:
                    crawler_config = json.load(f)
                cls.NAME_CONFIG_DICT[config_name] = crawler_config
            return func
        return wrapper



import datetime


class UserProfileBase(object):

    def __init__(self, username=None):
        self.username = username


class ContestBase(object):

    def __init__(self, contest_name=None, contest_begin_time=None):
        self.contest_begin_time = contest_begin_time

    def trans_to_date(self, date):
        """
            str date 转换成 datetime 格式
        """
        if isinstance(date, datetime.datetime):
            return date
        res_date = datetime.datetime.strptime(date, "%b/%d/%Y %H:%M")
        return res_date




import datetime
import time


def change_to_str(func):
    def wrapper(cls, message):
        try:
            func(cls, message)
            return func
        except:
            if not isinstance(message, (int, str)):
                message = message.__repr__()
            func(cls, message)
            return func
    return wrapper


class Logger:

    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BlUE = '\033[94m'
    END = '\033[0m'

    @classmethod
    @change_to_str
    def waring(cls, message):
        print("{}WARING:{}{}".format(cls.YELLOW, message, cls.END))

    @classmethod
    @change_to_str
    def error(cls, message):
        print("{}ERROR:{}{}".format(cls.RED, message, cls.END))
        cls.save_to_local(message)

    @classmethod
    @change_to_str
    def common(cls, message):
        print("{}{}{}".format(cls.GREEN, message, cls.END))

    @classmethod
    @change_to_str
    def save_to_local(cls, message):
        with open("../error_log.txt", "a") as f:
            now_time = datetime.datetime.now()
            f.write("{}:\n".format(datetime.datetime.strftime(now_time,'%Y-%m-%d %H:%M:%S')))
            f.write("{}\n".format(message))




from .base_crawler import CrawlerBase
from tools import register_config, Logger
from models import CodeforcesContestModel, CodeforcesUserInfoModel
from config import COMMON_ERROR

from bs4 import BeautifulSoup


class CodeforcesProfileCrawler(CrawlerBase):

    def __init__(self, username=None):
        self.username = username
        self._result_model = None # final result
        self.timer = None # may be extend timer[now not apply]
        self.profile_url = None
        self.contest_url = None
        self.headers = None
        self._preLoad()

    @register_config.ConfigBase.register(config_name="codeforces_profile")
    def _preLoad(self):
        config = register_config.ConfigBase.NAME_CONFIG_DICT["codeforces_profile"]
        if not config:
            Logger.error("codeforces config error!")
            return None
        self.profile_url = config["profile_url"]
        self.contest_url = config["contest_url"]
        self.headers = config["headers"]

    @staticmethod
    def _deal_with_contest_body(body=None):
        soup = BeautifulSoup(markup=body, features="lxml")
        result_set = soup.select("div.datatable>div>table>tbody>tr")
        last_contests_name = []
        latest_contests_ratings = []
        last_contest_time = []
        if len(result_set) > 10:
            result_set = result_set[:10]
        for item in result_set:
            contests = item.select("td")
            rating_t = (contests[3].string, contests[5].select_one("span").string, contests[6].string.strip('\r\n '))
            latest_contests_ratings.append(rating_t)
            last_contest_time.append(contests[2].text.strip('\r\n '))
            last_contests_name.append(contests[1].select_one("a").string.strip('\r\n '))
        return last_contests_name, latest_contests_ratings, last_contest_time

    @staticmethod
    def _deal_with_profile_body(body=None):
        soup = BeautifulSoup(markup=body, features="lxml")
        current_rating = soup.select_one("div.info>ul>li>span").string
        max_rating = soup.select("div.info>ul>li>span.smaller>span")[1].string
        problems_set = soup.select("div._UserActivityFrame_footer>div>div."
                                   "_UserActivityFrame_counter>div._UserActivityFrame_counterValue")
        last_month_solutions = problems_set[2].string
        solve_problems = problems_set[0].string
        Logger.waring(solve_problems)
        return max_rating, current_rating, last_month_solutions, solve_problems

    def _get_model(self) -> CodeforcesUserInfoModel:
        # error judge
        profile_body = self.get_request_body(url=self.profile_url,
                                             headers=self.headers, username=self.username)
        if not self.check_result(profile_body):
            return CodeforcesUserInfoModel(err_msg=profile_body)
        contest_body = self.get_request_body(url=self.contest_url,
                                             headers=self.headers, username=self.username)
        if not self.check_result(contest_body):
            return CodeforcesUserInfoModel(err_msg=contest_body)

        if profile_body is None or contest_body is None:
            Logger.error("variable is None")
            return COMMON_ERROR
        # error judge
        max_rating, current_rating, last_month_solutions, solve_problems = \
            self._deal_with_profile_body(body=profile_body)
        last_contests_name, latest_contests_ratings, last_contest_time = \
            self._deal_with_contest_body(body=contest_body)
        result_model = CodeforcesUserInfoModel(
            username=self.username,
            max_rating=max_rating,
            current_rating=current_rating,
            solve_problems=solve_problems,
            last_month_solutions=last_month_solutions,
            last_contest_time=last_contest_time,
            latest_contests_ratings=latest_contests_ratings,
            last_contests_name=last_contests_name,
        )
        return result_model

    @property
    def result_info(self) -> str:
        if self._result_model is None:
            self._result_model = self._get_model()
        result_info = self._result_model.__repr__()
        return result_info

    @property
    def result_model(self):
        if self._result_model is None:
            self._result_model = self._get_model()
        return self._result_model


class CodeforcesContestCrawler(CrawlerBase):

    def __init__(self, username=None):
        self.username = username
        self.headers = None
        self.profile_url = None
        self._preLoad()
        self._result_model = None

    @register_config.ConfigBase.register(config_name="codeforces_contest")
    def _preLoad(self):
        config = register_config.ConfigBase.NAME_CONFIG_DICT["codeforces_contest"]
        if not config:
            Logger.error("codeforces_crawler.py line 97 : codeforces config error!")
            return None
        self.headers = config["headers"]
        self.profile_url = config["profile_url"]

    @staticmethod
    def deal_with_contest_body(contest_body):
        contest_name_list = []
        contest_start_time_list = []
        contest_length_list = []
        contest_url_list = []
        soup = BeautifulSoup(markup=contest_body, features="lxml")
        contest_list = soup.select_one("div.datatable>div>table")
        contest_list = contest_list.select("tr")
        for contest in contest_list[1:]:
            contest_url_list.append("https://codeforces.com/contest/{}"\
                                    .format(contest['data-contestid'].strip('\n\r ')))
            items = contest.select("td")
            contest_name_list.append(items[0].string.strip('\n\r '))
            contest_start_time_list.append(items[2].select_one("span").string.strip('\n\r '))
            contest_length_list.append(items[3].string.strip('\n\r '))

        return contest_name_list, contest_start_time_list\
            , contest_length_list, contest_url_list

    def _get_model(self) -> CodeforcesContestModel:
        contest_body = self.get_request_body(url=self.profile_url, headers=self.headers)
        if not self.check_result(contest_body):
            return CodeforcesContestModel(err_msg=contest_body)
        if contest_body is None:
            return CodeforcesContestModel(err_msg=COMMON_ERROR)
        contest_name_list, contest_start_time_list, contest_length_list, contest_url_list = \
            self.deal_with_contest_body(contest_body=contest_body)
        res_model = CodeforcesContestModel(
            contest_name_list=contest_name_list,
            contest_start_time_list=contest_start_time_list,
            contest_length_list=contest_length_list,
            contest_url_list=contest_url_list
        )
        return res_model

    @property
    def result_model(self):
        if self._result_model is None:
            self._result_model = self._get_model()
        return self._result_model


if __name__ == '__main__':
    # 测试
    crawler = CodeforcesContestCrawler()
    model = crawler.result_model
    print(model.today_contests_string())
    crawler_ = CodeforcesProfileCrawler(username="CCoolGuang")
    model = crawler_.result_info
    print(model)


from .base_model import UserProfileBase, ContestBase

import datetime


class CodeforcesUserInfoModel(UserProfileBase):

    def __init__(self,
                 username=None,
                 max_rating=None,
                 current_rating=None,
                 solve_problems=None,
                 last_month_solutions=None,
                 last_contest_time=None,
                 latest_contests_ratings=None,
                 last_contests_name=None,
                 err_msg=None
                 ):
        self.username = username
        self.max_rating = max_rating
        self.current_rating = current_rating
        self.solve_problems = solve_problems
        self.last_month_solutions = last_month_solutions
        # 最近比赛（<10）的时间
        self.last_contest_time = last_contest_time
        # 元组形式分别为（rank，change，new_rating）
        self.latest_contests_ratings = latest_contests_ratings
        # 最近比赛（<10）的名称
        self.last_contests_name = last_contests_name
        self.err_msg = err_msg

    def __repr__(self):
        if self.err_msg:
            return self.err_msg
        return "----(Codeforces个人信息)----" \
               "\n用户名: {}" \
               "\n最高Rating:{}" \
               "\n当前Rating:{}" \
               "\n最近一场rating波动:{}" \
               "\n最近一场比赛rank:{}" \
               "\n最近一场比赛时间:{}" \
               "\n最近一场比赛名称:{}" \
               "\n最近十场比赛平均rating:{}" \
               "\n总解决题目数量:{}" \
               "\n上个月解决题目数量:{}".format(self.username, self.max_rating, self.current_rating, self.latest_contest_change,
                                       self.latest_contests_rank, self.latest_contest_time, self.latest_contest_name,
                                       self.latest_avg_contests_rating, self.solve_problems, self.last_month_solutions)

    @property
    def latest_avg_contests_rating(self):
        """
            最近（<10）场比赛的平均rating
        """
        avg = 0
        if len(self.latest_contests_ratings) == 0:
            return None
        for item in self.latest_contests_ratings:
            avg += int(item[2])
        return avg / len(self.latest_contests_ratings)

    @property
    def latest_contest_time(self):
        """
            最近一场比赛的时间
            type : str
        """
        if len(self.last_contest_time) == 0:
            return None
        return self.last_contest_time[0]

    @property
    def latest_contest_name(self):
        """
            最近一场比赛的名称
            type : str
        """
        if len(self.last_contests_name) == 0:
            return None
        return self.last_contests_name[0]

    @property
    def latest_contest_change(self):
        """
            最近1 场比赛的rating change
            return type : str
        """
        if len(self.latest_contests_ratings) == 0:
            return None
        return self.latest_contests_ratings[0][1]

    @property
    def late_contests_change(self):
        """
            最近（<10）场比赛的rating change
            type : tuple
        """
        if len(self.latest_contests_ratings) == 0:
            return None
        result = (item[1] for item in self.latest_contests_ratings)
        return result

    @property
    def late_contests_ratings(self):
        """
            最近（<10）场比赛的rating
            type : tuple
        """
        if len(self.latest_contests_ratings) == 0:
            return None
        result = (item[2] for item in self.latest_contests_ratings)
        return result

    @property
    def late_contests_rank(self):
        """
            最近比赛的(<10)rank排名
            type : tuple
        """
        if len(self.latest_contests_ratings) == 0:
            return None
        result = (item[0] for item in self.latest_contests_ratings)
        return result

    @property
    def latest_contests_rank(self):
        """
            最近1场比赛的rank排名
            return type: str
        """
        if len(self.latest_contests_ratings) == 0:
            return None
        return self.latest_contests_ratings[0][0]


class CodeforcesContestModel(ContestBase):

    def __init__(self, contest_name_list=None, contest_start_time_list=None, contest_length_list=None, \
                 contest_url_list=None, err_msg=None):
        self.contest_name_list = contest_name_list
        self.contest_start_time_list = contest_start_time_list
        self.contest_length_list = contest_length_list
        self.contest_url_list = contest_url_list
        # self.contest_info -> (名称，开始时间，持续时常，比赛链接）
        try:
            self.contests_info = [(x, y, z, u) for x, y, z, u in zip(contest_name_list, contest_start_time_list,
                                                                     contest_length_list, contest_url_list)]
        except TypeError:
            self.contests_info = []
        self.err_msg = err_msg

    @property
    def today_contests(self):
        """
            获取今天的比赛信息
            :return -> tuple list [(name, time. length)]
        """
        _now = self.trans_to_date(datetime.datetime.now())
        contests = []
        try:
            for date_item in self.contests_info:
                trans_date_item = self.trans_to_date(date_item[1])
                if trans_date_item.day == _now.day and trans_date_item.month == _now.month \
                        and trans_date_item.year == _now.year:
                    contests.append(date_item)
        except TypeError:
            pass
        return contests

    def today_contests_string(self):
        if self.err_msg:
            return self.err_msg
        return self._get_contests_string(contest_list=self.today_contests)

    def recent_contest(self):
        if self.err_msg:
            return self.err_msg
        return self.contests_info

    def recent_contests_string(self):
        if self.err_msg:
            return self.err_msg
        return self._get_contests_string(contest_list=self.contests_info)

    def _get_contests_string(self, contest_list=None):
        result = ""
        if len(contest_list) == 0:
            result = "暂无任何比赛信息"
            return result
        for item in contest_list:
            result += "=========================\n"
            result += "比赛名称:{}\n" \
                      "开始时间:{}\n" \
                      "比赛时长:{}\n" \
                      "比赛链接:{}\n".format(item[0], item[1],
                                         item[2], item[3])
        return result
